
#import "MiniSQLAdaptor.h"
#import "MiniSQLAdaptorContext.h"
#import "MiniSQLAdaptorChannel.h"
#import "MiniSQLExpression.h"

#import <foundation/NSArray.h>
#import <foundation/NSString.h>
#import <foundation/NSBundle.h>
#import <foundation/NSDictionary.h>
#import <foundation/NSUtilities.h>
#import <foundation/NSString.h>
#import <foundation/NSValue.h>

#import <eoaccess/EONull.h>
#import <eoaccess/EOAttribute.h>
#import <eoaccess/EOQuotedExpression.h>

@implementation MiniSQLAdaptor

- (Class)expressionClass
// Returns the subclass of EOSQLExpression used by this adaptor for query
// language expressions.  You directly invoke this method, but an
// adaptor subclass must properly implement this method.
{
    return [MiniSQLExpression class];
}

- (EOAdaptorContext *)createAdaptorContext
// Returns a new EOAdaptorContext, or nil if a new context cannot be
// created.  EOAdaptors by default have no contexts at all.  The newly
// created context retains its adaptor.
{
    return [[[MiniSQLAdaptorContext alloc] initWithAdaptor:self] autorelease];
}


- (BOOL) hasOpenChannels
// Returns YES if the adaptor has any open channels, NO otherwise. See
// -openChannel in EOAdaptorChannel.h.
{
    return (openChannels > 0) ? YES : NO;
}

- (void) miniSQLChannelDidConnect
{
    openChannels += 1;
}

- (void)miniSQLChannelDidDisconnect
{
    if (openChannels > 0) {
	openChannels -= 1;
    }
}

- (BOOL)hasValidConnectionDictionary
// Returns YES if the adaptor can connect with the current connection
// dictionary, NO otherwise. An actual connection is made when the first
// adaptor channel is sent an -openChannel message.
{
    if ([self hasOpenChannels]) {
	return YES;
    } else {
	MiniSQLAdaptorChannel *aChannel = [[[MiniSQLAdaptorChannel alloc]
					    initWithAdaptor:self
					    andContext:nil]
					    autorelease];

	return [aChannel connectWithConnectionDictionary:[self connectionDictionary]];
    }
}

- formatAttribute:(EOAttribute *)attribute
{
    return attribute;
}

- formatValue:value forAttribute:(EOAttribute *)attribute
{
    if (value == nil || 
    	value == [EONull null]) {
	return @"NULL";
    } else {
    	NSString *type = [attribute externalType];
	
	if ([type isEqualToString:@"INT"]) {
	    return value;
	} else if ([type isEqualToString:@"REAL"]) {
	    NSNumber           *number = value;
	    NSString           *stringValue = [number stringValue];

	// MiniSQL doesn't understand REAL values specified
	// without a decimal point, so we add one if the float
	// doesn't itself...
	    if ([stringValue rangeOfString:@"."].length == 0) {
		return [NSString stringWithFormat:@"%@.0", stringValue];
	    } else {
		return stringValue;
	    }
	} else {
	    return [[[EOQuotedExpression alloc]
		    initWithExpression:value
		    quote:@"'"
		    escape:@"'"]
		    autorelease];
	}
    }
}

- (BOOL) isValidQualifierType:(NSString *)typeName
{
 // everything is permitted...
    return YES;
}
@end
