#import <eoaccess/EOAdaptorContext.h>

@class MiniSQLAdaptor;
@class MiniSQLAdaptorChannel;

@interface MiniSQLAdaptorContext:EOAdaptorContext
{
    MiniSQLAdaptor		*adaptor;
    MiniSQLAdaptorChannel	*channel;

    int				transaction;
}

- initWithAdaptor:(MiniSQLAdaptor *)anAdaptor;
- (void)dealloc;

- (EOAdaptor *)adaptor;
// Returns the EOAdaptor that handles the connection within which the
// adaptor context works.

- (EOAdaptorChannel *)createAdaptorChannel;
// Returns a new EOAdaptorChannel.  Returns nil if a new channel cannot
// be created.  EOAdaptorContexts by default have no channels at all.  The
// newly created channel retains its context.

- (BOOL)beginTransaction;
// Attempts to begin a new transaction, nested within the current one if
// nested transaction are supported, returning YES if successful and NO
// if not.

- (BOOL)commitTransaction;
// Attempts to commit the last transaction begun, returning YES if
// successful and NO if not.

- (BOOL)rollbackTransaction;
// Attempts to roll back the last transaction begun, returning YES if
// successful and NO if not.

- (BOOL)canNestTransactions;
// Returns YES if the database server can nest transactions, NO
// otherwise.

- (unsigned)nestedTransactions;
// Returns the number of transactions in progress. If transactions
// can nest, this number may be greater than one.

- (unsigned)transactionNestingLevel;
// A cover for -nestedTransactions. 

- (BOOL)hasBusyChannels;
// Returns YES if the receiver has channels that have outstanding
// operations (that is, have a fetch in progress), NO otherwise.

@end
