#import "MiniSQLExpression.h"

#import <eoaccess/EOQualifier.h>
#import <eoaccess/EOEntity.h>
#import <eoaccess/EOAttribute.h>
#import <eoaccess/EORelationship.h>
#import <eoaccess/EOExpressionArray.h>
#import <eoaccess/EOAttributeOrdering.h>

#import <foundation/NSUtilities.h>
#import <foundation/NSCharacterSet.h>

@implementation MiniSQLExpression:EOSQLExpression

- (NSString *)expressionValueForAttribute:(EOAttribute *)anAttribute;
    // Does the appropriate aliasing and name generation for an attribute in
    // the context of this particular query.
{
    if ([anAttribute isFlattened]) {
	NSString           *component;
	NSArray            *components = [[anAttribute definition] componentsSeparatedByString:@"."];
	NSEnumerator       *enumerator = [components objectEnumerator];

	EOEntity		*entity = [anAttribute entity];
	EORelationship		*relationship;
	
	for (entity = [anAttribute entity];
	     (component = [enumerator nextObject]) &&
	     (relationship = [entity relationshipNamed:component]);
	     entity = [relationship destinationEntity]) {
	}
	
	return [self expressionValueForAttribute:[entity attributeNamed:component]];
     } else {
	return [NSString stringWithFormat:@"%@.%@",
	    [[anAttribute entity] externalName],
	    [anAttribute columnName]];
    }
}


- (NSMutableDictionary *)relationshipPathsForAttributes:(NSArray *)attributes
    qualifier:(EOQualifier *)qualifier fetchOrder:(NSArray *)fetchOrder
    // The keys in the returned dictionary represent paths to relationships
    // referenced by this expression and their associated objects are aliases
    // that can be used in the query language.
{
    EOAttribute  *anAttribute;
    NSEnumerator *enumerator0 = [attributes objectEnumerator];
    
    NSMutableDictionary *relationshipPaths =
	[NSMutableDictionary dictionaryWithCapacity:[attributes count]]; 

    while ((anAttribute = [enumerator0 nextObject]) != nil) {
	if ([anAttribute isFlattened]) {
	    NSString           *component;	    
	    NSArray            *components = [[anAttribute definition] componentsSeparatedByString:@"."];
	    NSEnumerator       *enumerator = [components objectEnumerator];
    
	    EOEntity		*entity = [anAttribute entity];
	    EORelationship		*relationship;
	    
	    for (entity = [anAttribute entity];
		    (component = [enumerator nextObject]) &&
		    (relationship = [entity relationshipNamed:component]);
		    entity = [relationship destinationEntity]) {
	    }
	    
	    [relationshipPaths setObject:[entity externalName]
		forKey:[components objectAtIndex:0]];
	} else {
	    [relationshipPaths setObject:[[anAttribute entity] externalName]
	    forKey:@""];
	}
    }
    
    return relationshipPaths;
}


- fromClauseForRelationshipPaths:(NSDictionary *)relationships
    entity:(EOEntity *)entity
    // Builds a SQL FROM clause using a dictionary of aliased relation paths
{
    id                  alias;
    NSEnumerator       *enumerator = [relationships objectEnumerator];
    
    NSString		*entityName = [entity externalName];
    
    EOExpressionArray *array = [[[EOExpressionArray alloc]
    	initWithPrefix:@""
	infix:@", "
	suffix:@""]
	    autorelease];

    [array addObject:entityName];

    while ((alias = [enumerator nextObject]) != nil) {
    	if ([alias length] && ![alias isEqualToString:entityName]) {
	    [array addObject:alias];
	}
    }

    return [NSString stringWithFormat:@"FROM %@", [array expressionValueForContext:array]];
}


- whereClauseForQualifier:(EOQualifier *) qualifier
    joinExpresssion:joinExpression
 // Builds an SQL WHERE clause using the qualifier and joinExpression 
{
    NSMutableString    *newExpression =
	[NSMutableString stringWithCapacity:80];

    NSMutableString    *newQualification =
	[[qualifier expressionValueForContext:self] mutableCopy];

    NSMutableString    *newJoinExpression =
	[[joinExpression expressionValueForContext:self] mutableCopy];


    if (qualifier) {
	[newQualification deleteCharactersInRange:
	    [newQualification rangeOfString:@"("]];
	[newQualification deleteCharactersInRange:
	    [newQualification rangeOfString:@")"]];
	[newExpression appendString:newQualification];
    }
    if (joinExpression) {
	[newJoinExpression deleteCharactersInRange:
	    [newJoinExpression rangeOfString:@"("]];
	[newJoinExpression deleteCharactersInRange:
	    [newJoinExpression rangeOfString:@")"]];

	if ([newQualification length] > 0) {
	    [newExpression appendString:@" AND "];
	}
	[newExpression appendString:newJoinExpression];
    }
    if ([newExpression length] > 0) {
	return[NSString stringWithFormat:@"WHERE %@", newExpression];
    } else {
	return @"";
    }
}

@end

